"""
Methods for using the git log command
"""
import re
import subprocess
from datetime import datetime
from pathlib import Path
from typing import Optional

from .constants import EMPTY_REPO_RE, UNKNOWN_REV_RE
from .datatypes import Log
from .exceptions import (GitException, NoCommitsException, NoLogsException,
                         UnknownRevisionException)

__all__ = ["get_logs"]


def __process_log(stdout_line: str):
    parts = stdout_line.split(";;")
    if len(parts) != 4:
        raise ValueError("invalid log line: {stdout_line}")
    return Log(
        parts[0],
        parts[1],
        datetime.fromisoformat(parts[2]),
        parts[3]
    )


def __process_logs(stdout: str):
    log_lines = stdout.strip().split("\n")
    return map(__process_log, log_lines)


def get_logs(
        git_repo: Path,
        branch: Optional[str] = None,
        max_number: Optional[int] = None,
        since: Optional[datetime] = None,
        until: Optional[datetime] = None):
    """
    Generate git logs from a repo

        :param git_repo: Path to the repo
        :param branch: The branch name, defaults to None
        :param max_number: max number of logs to get, defaults to None
        :param since: Filter logs after given date, defaults to None
        :param until: Filter logs before given date defaults to None
        :raises NoCommitsException: Repo has no commits
        :raises UnknownRevisionException: Unknown revision/branch name
        :raises GitException: Error to do with git
        :raises NoLogsException: No logs have been generated
        :return: The generated logs
    """
    args = ["git", "-C", str(git_repo), "log"]

    if branch is not None:
        args.append(str(branch))
    if max_number is not None:
        args.append(f"--max-count={max_number}")
    if since is not None:
        args.append(f"--since={since.isoformat()}")
    if until is not None:
        args.append(f"--until={until.isoformat()}")
    args.append("--pretty=%H;;%ae;;%cI;;%s")

    process_status = subprocess.run(
        args,
        capture_output=True)
    if not process_status.stdout:
        stderr = process_status.stderr.decode()
        if re.match(EMPTY_REPO_RE, stderr):
            raise NoCommitsException()
        if re.match(UNKNOWN_REV_RE, stderr):
            raise UnknownRevisionException(f"unknown revision/branch {branch}")
        if process_status.returncode != 0:
            raise GitException(stderr)
        raise NoLogsException(f"no logs found (using given filters) for '{git_repo.name}'")
    stdout = process_status.stdout.decode()
    return __process_logs(stdout)



"""
Methods that don't fit in their own file
"""
import subprocess
from pathlib import Path
from typing import Optional

from .datatypes import ArchiveTypes
from .exceptions import AlreadyExistsException, GitException

__all__ = [
    "init_repo", "get_description",
    "set_description", "run_maintenance",
    "get_archive",
]


def init_repo(
        repo_dir: Path,
        repo_name: str,
        bare: bool = True,
        default_branch: Optional[str] = None):
    """
    Creates a new git repo in the directory with the given name,
    if bare the repo name will have .git added at the end.

        :param repo_dir: Where the repo will be
        :param repo_name: The name of the repo
        :param bare: Whether the repo is bare, defaults to True
        :param default_branch: The branch name to use, defaults to None
        :raises AlreadyExistsException: A repo already exists
        :raises GitException: Error to do with git
    """
    if bare:
        repo_name = repo_name + ".git"
    repo_path = repo_dir / repo_name

    if (repo_path.exists()):
        raise AlreadyExistsException(f"path already exists for '{repo_name}'")

    args = ["git", "init", str(repo_path), "--quiet"]
    if bare:
        args.append("--bare")
    if default_branch:
        args.append(f"--initial-branch={default_branch}")
    process = subprocess.run(args)
    if process.returncode != 0:
        raise GitException(process.stderr.decode())


def get_description(git_repo: Path) -> str:
    """
    Gets the set description for a repo

        :param git_repo: Path to the repo
        :return: The description
    """
    with open(git_repo / "description", "r") as fo:
        return fo.read()


def set_description(git_repo: Path, description: str):
    """
    Sets the set description for a repo

        :param git_repo: Path to the repo
    """
    with open(git_repo / "description", "w") as fo:
        fo.write(description)


def run_maintenance(git_repo: Path):
    """
    Run a maintenance git command to specified repo

        :param git_repo: Where the repo is
        :raises GitException: Error to do with git
    """
    args = ["git", "-C", str(git_repo), "maintenance", "run"]
    process = subprocess.run(args, capture_output=True)
    if process.returncode != 0:
        raise GitException(process.stderr.decode())


def get_archive(
        git_repo: Path,
        archive_type: ArchiveTypes,
        tree_ish: str = "HEAD") -> bytes:
    """
    get a archive of a git repo

        :param git_repo: Where the repo is
        :param archive_type: What archive type will be created
        :param tree_ish: What commit/branch to save, defaults to "HEAD"
        :raises GitException: Error to do with git
        :return: The content of the archive ready to write to a file
    """
    # this allows for strings to be passed
    if isinstance(archive_type, ArchiveTypes):
        archive_type = archive_type.value
    process = subprocess.run(
        ["git", "-C", str(git_repo), "archive", f"--format={archive_type}", tree_ish],
        capture_output=True)
    if process.returncode != 0:
        raise GitException(process.stderr.decode())
    return process.stdout



"""
Exceptions that could be raised during one of the git commands
"""


class GitException(Exception):
    pass


class NoCommitsException(GitException):
    pass


class NoBranchesException(GitException):
    pass


class AlreadyExistsException(GitException):
    pass


class NoLogsException(GitException):
    pass


class UnknownRevisionException(GitException):
    pass


class UnknownRefException(GitException):
    pass
